/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include "jabber.h"
#include "jabprivate.h"
#include "jabstream.h"
#include "log.h"

void _jabstream_main(jabc j, void *buffer, int bufsz);


void jabstream_write(jabc j, char *buf)
{
#ifdef DEBUG
	jdebug(ZONE,"\nwrite[%s]",buf);
#endif
	write(jab_fd(j),buf,strlen(buf));
}


void jabstream_read(jabc j)
{
	static char buf[4096];
	int sz;

	sz = read(jab_fd(j),buf,4095);
	if (sz > 0) {
		buf[sz] = 0;
#ifdef DEBUG
		jdebug(ZONE,"read[%s]",buf);
#endif
		_jabstream_main(j,buf,sz);
		return;
	}

	_jab_parsenode(j, NULL, JABSTREAM_SOCKERR);
}


void jabstream_start(jabc j, xmlnode x)
{
	spool s;
	char *head, *fixr;

	if(xmlnode_has_children(x))
		return;

	s = spool_new(x->p);
	spooler(s,"<?xml version=\'1.0\'?>",xmlnode2str(x),s);
	head = spool_print(s);

    /* xmlnode is acting weirder and weirder. The following handles both situations. */
    fixr = strstr(head, "></");
	if(!fixr)
	{
    	fixr = strstr(head,"/>");
    	*fixr = '>';
    }
	
	++fixr;
	*fixr = '\0';

	write(jab_fd(j), head, strlen(head));
	xmlnode_free(x);

	_jabstream_main(j,0,0);
}


xmlnode jabstream_header(char *xmlns, char *to, char *from)
{
	xmlnode x;

	x = xmlnode_new_tag("stream:stream");
	xmlnode_put_attrib(x, "xmlns:stream", "http://etherx.jabber.org/streams");
	if(xmlns) xmlnode_put_attrib(x, "xmlns", xmlns);
	if(to) xmlnode_put_attrib(x, "to", to);
	if(from) xmlnode_put_attrib(x, "from", from);
	return x;
}


void jabstream_stop(jabc j, int f, char *err)
{
	if (JAB_S_OFF == j->state) return;
	j->state = JAB_S_OFF;

	if (f) {
		int s = jab_fd(j);
		if(err) {
			write(s, "<stream:error>", 14);
			write(s, err, strlen(err));
			write(s, "</stream:error>", 15);
		}
		write(s,"</stream:stream>",16);
	}
	XML_ParserFree(j->xparser);
	close(jab_fd(j));
}


/*
** private functions for handling xml-xmlnode conversion via expat
*/

void _jabstream_startElement(void* userdata, const char* name, const char** atts)
{
	jabc j = (jabc)userdata;
	xmlnode current = j->xnode;

	if(current == NULL)
	{
		current = xmlnode_new_tag(name);
		xmlnode_put_expat_attribs(current, atts);

		if(j->xflag == 0)
		{ /* this is a special flag saying that this is the <stream> root */
			j->xflag = 1;
			_jab_parsenode(j, current, JABSTREAM_ROOT);
			xmlnode_free(current);
			current = NULL;
		}
	}
	else
	{
		current = xmlnode_insert_tag(current, name);
		xmlnode_put_expat_attribs(current, atts);
	}

	j->xnode = current;
}


void _jabstream_endElement(void* userdata, const char* name)
{
	jabc j = (jabc)userdata;
	xmlnode current = j->xnode;
	xmlnode parent;

	/* if it's already NULL we've received </stream>, tell the app and we're outta here */
	if(current == NULL)
	{
		_jab_parsenode(j, current, JABSTREAM_CLOSE);
	} else {
		parent = xmlnode_get_parent(current);
		if(parent == NULL)
		{ /* we are the top-most node, feed to the app and delete */
		_jab_parsenode(j, current, JABSTREAM_NODE);
		xmlnode_free(current);
		}
		j->xnode = parent;
	}
}


void _jabstream_charData(void* userdata, const char* s, int len)
{
	jabc j = (jabc)userdata;
	xmlnode current = j->xnode;

	if(current == NULL)
	{
		/* we must be in the root of the stream where CDATA is irrelevant */
		return;
	}

	xmlnode_insert_cdata(current, s, len);
}


void _jabstream_main(jabc j, void *buffer, int bufsz)
{
	XML_Parser p = j->xparser;

	if(buffer == NULL)
	{
		p = XML_ParserCreate(NULL);
		j->xparser = p;
		XML_SetUserData(p, (void *)j);
		XML_SetElementHandler(p, (void *)_jabstream_startElement, (void *)_jabstream_endElement);
		XML_SetCharacterDataHandler(p, (void *)_jabstream_charData);
		return;
	}

	if(!XML_Parse(p, buffer, bufsz, 0))
	{
		/* error("jabstream_parse_error: %s",(char *)XML_ErrorString(XML_GetErrorCode(p))); */
		_jab_parsenode(j, NULL, JABSTREAM_XMLERR);
	}
}
